#include "../Framework2/OpenGL/OpenGLApp.h"
#include "../Framework2/Util/OpenGLModel.h"
#include "../Framework2/Util/CollisionSet.h"

#include "../Framework2/Util/TexturePacker.h"

#define SAMPLE_COUNT 400
#define LIGHTMAP_COUNT 16

vec3 getLightPos(float zp){
	float z = zp;
	float y = sqrtf(1.5f - z * z);

	y *= 600;
	z *= 600;

	return vec3(256, 416 - y, z);
}

struct Vertex {
	vec3 pos;
	vec2 texCoord;
	vec3 tangent;
	vec3 binormal;
	vec3 normal;
	vec2 texCoord2;
};

class MainApp : public OpenGLApp {
public:
	void processMovement(const vec3 &newPosition);
	void resetCamera();

	bool init();
	bool exit();

	bool load();
	bool unload();

	void drawLight();
	bool drawFrame();
protected:
	CollisionSet coll;

	vec3 lightPos;
	float lightSize;

	ShaderID lighting;
	OpenGLModel *model;

	TextureID base[4], bump[4], lightMaps[LIGHTMAP_COUNT];
	TextureID light;
};

void MainApp::processMovement(const vec3 &newPosition){
	vec3 pos = newPosition;

	coll.pushSphere(pos, 48);

	position = pos;
}

void MainApp::resetCamera(){
	position = vec3(-768, 0, -1152);
	wx = 0.1f;
	wy = -0.75f;
	wz = 0;
}

struct PolyData {
	vec3 v0, vDir0, vDir1;
	vec2 t0, tDir0, tDir1;
};

bool MainApp::init(){
	unsigned int i, j;

	model = new OpenGLModel();
	model->loadFromFile("../Models/Room3/Map.hmdl");

	for (i = 0; i < model->getBatchCount(); i++){
		model->getBatch(i)->insertAttribute(ATT_TEXCOORD, ATT_FLOAT, 2, 4);
	}

	TexturePacker texPacker;
	for (i = 0; i < model->getBatchCount(); i++){
		Batch *batch = model->getBatch(i);
		Vertex *src = (Vertex *) batch->getVertices();
		unsigned short *inds = (unsigned short *) batch->getIndices();

		for (j = 0; j < batch->getIndexCount(); j += 6){
			float w = length(src[inds[j + 1]].pos - src[inds[j]].pos) / 34;
			float h = length(src[inds[j + 5]].pos - src[inds[j]].pos) / 34;

			if (w < 8) w = 8;
			if (h < 8) h = 8;

			unsigned int tw = (int) w;
			unsigned int th = (int) h;

			tw = (tw + 4) & ~7;
			th = (th + 4) & ~7;

			texPacker.addRectangle(tw, th);
		}
	}

	unsigned int tw = 256;
	unsigned int th = 256;

	if (!texPacker.assignCoords(&tw, &th)){
#ifdef _WIN32
		MessageBox(hwnd, "Lightmap too small", "Error", MB_OK);
#endif
		return false;
	}


	tw = getUpperPowerOfTwo(tw);
	th = getUpperPowerOfTwo(th);


	Set <PolyData *> polyData;

	unsigned int index = 0;
	for (i = 0; i < model->getBatchCount(); i++){
		Batch *batch = model->getBatch(i);
		Vertex *src = (Vertex *) batch->getVertices();
		unsigned short *inds = (unsigned short *) batch->getIndices();

		for (j = 0; j < batch->getIndexCount(); j += 6){
			TextureRectangle *rect = texPacker.getRectangle(index);

			float x0 = float(rect->x + 0.5f) / tw;
			float x1 = float(rect->x + rect->width - 0.5f) / tw;
			float y0 = float(rect->y + 0.5f) / th;
			float y1 = float(rect->y + rect->height - 0.5f) / th;

			src[inds[j    ]].texCoord2 = vec2(x0, y0);
			src[inds[j + 1]].texCoord2 = vec2(x1, y0);
			src[inds[j + 2]].texCoord2 = vec2(x1, y1);
			src[inds[j + 5]].texCoord2 = vec2(x0, y1);

			PolyData *pd = new PolyData;
			pd->v0 = src[inds[j]].pos;
			pd->vDir0 = src[inds[j + 1]].pos - pd->v0;
			pd->vDir1 = src[inds[j + 5]].pos - pd->v0;
			pd->t0 = vec2(x0, y0);
			pd->tDir0 = vec2(x1 - x0, 0);
			pd->tDir1 = vec2(0, y1 - y0);
			/*int ind = */polyData.add(pd);
			//PolyData *poly = polyData.getArray() + ind;

			coll.addTriangle(src[inds[j]].pos, src[inds[j + 1]].pos, src[inds[j + 2]].pos, pd);
			coll.addTriangle(src[inds[j + 3]].pos, src[inds[j + 4]].pos, src[inds[j + 5]].pos, pd);

			index++;
		}
	}
	coll.build(3, 1, 50);

	lightSize = 180;


	/*****	Lightmap generation code *****/

/*
	vec3 samples[SAMPLE_COUNT];
	for (unsigned int x = 0; x < SAMPLE_COUNT; x++){
		vec3 v;
		do {
			v = vec3(float(rand()), float(rand()), float(rand())) * (2.0f / RAND_MAX) - 1.0f;
		} while (dot(v, v) > 1);
		samples[x] = v * lightSize;
	}


	for (unsigned int lm = 0; lm < LIGHTMAP_COUNT; lm++){

		float zp = float(lm) / (LIGHTMAP_COUNT - 1) * 2 - 1;

		lightPos = getLightPos(zp);


		unsigned char *lMap = new unsigned char[2 * tw * th];
		memset(lMap, 0, tw * th);

		// Pass 1, illuminate using the main light
		index = 0;
		for (i = 0; i < model->getBatchCount(); i++){
			Batch *batch = model->getBatch(i);
			Vertex *src = (Vertex *) batch->getVertices();
			unsigned short *inds = (unsigned short *) batch->getIndices();

			for (j = 0; j < batch->getIndexCount(); j += 6){
				TextureRectangle *rect = texPacker.getRectangle(index);

				vec3 dirS = (src[inds[j + 1]].pos - src[inds[j]].pos);
				vec3 dirT = (src[inds[j + 5]].pos - src[inds[j]].pos);

				vec3 pos = src[inds[j]].pos;
				vec3 normal = src[inds[j]].normal;

				for (int t = 0; t < int(rect->height); t++){
					for (int s = 0; s < int(rect->width); s++){
						vec3 samplePos = pos + saturate((float(s) / (rect->width - 1))) * dirS + saturate((float(t) / (rect->height - 1))) * dirT;

						float light = 0.0f;

						vec3 lightVec = lightPos - samplePos;
						float diffuse = dot(normalize(lightVec), normal);

						if (diffuse > 0){
							coll.pushSphere(samplePos, 5);

							for (unsigned int k = 0; k < SAMPLE_COUNT; k++){
								if (!coll.intersect(samplePos, lightPos + samples[k])) light += 1.0f / SAMPLE_COUNT;
							}

							light *= saturate(diffuse) / (1.0f + 0.0000004f * dot(lightVec, lightVec));
						}

						lMap[((rect->y + t) * tw + (rect->x + s))] = (unsigned char) (255.0f * light);
						//lMap[2 * ((rect->y + t) * tw + (rect->x + s))] = (unsigned char) (255.0f * light);
					}
				}

				index++;
			}
		}
*/

/*
		// Pass 2, indirect lumination
		index = 0;
		for (i = 0; i < model->getBatchCount(); i++){
			Batch *batch = model->getBatch(i);
			Vertex *src = (Vertex *) batch->getVertices();
			unsigned short *inds = (unsigned short *) batch->getIndices();

			for (j = 0; j < batch->getIndexCount(); j += 6){
				TextureRectangle *rect = texPacker.getRectangle(index);

				vec3 dirS = (src[inds[j + 1]].pos - src[inds[j]].pos);
				vec3 dirT = (src[inds[j + 5]].pos - src[inds[j]].pos);

				vec3 pos = src[inds[j]].pos;
				vec3 normal = src[inds[j]].normal;

				for (int t = 0; t < int(rect->height); t++){
					for (int s = 0; s < int(rect->width); s++){
						vec3 samplePos = pos + saturate((float(s) / (rect->width - 1))) * dirS + saturate((float(t) / (rect->height - 1))) * dirT;

						coll.pushSphere(samplePos, 5);

						float light = 0.0f;
						for (unsigned int k = 0; k < SAMPLE_COUNT; k++){
							vec3 point;
							PolyData *polyData;
							if (coll.intersect(samplePos, samplePos + 100 * samples[k], &point, true, (void **) &polyData)){
								float diffuse = dot(normalize(point - samplePos), normal);

								if (diffuse > 0){
									vec3 dir = point - polyData->v0;
									float td0 = dot(dir, polyData->vDir0) / dot(polyData->vDir0, polyData->vDir0);
									float td1 = dot(dir, polyData->vDir1) / dot(polyData->vDir1, polyData->vDir1);

									vec2 coord = polyData->t0 + td0 * polyData->tDir0 + td1 * polyData->tDir1;

									int x = int(coord.x * tw + 0.5);
									int y = int(coord.y * th + 0.5);

									if (x < 0) x = 0;
									if (x >= (int) tw) x = tw - 1;
									if (y < 0) y = 0;
									if (y >= (int) th) y = th - 1;

									light += diffuse * float(lMap[2 * (y * tw + x)]) / SAMPLE_COUNT;

									//light += 0.1f / SAMPLE_COUNT;
								}
							}
						}

						lMap[2 * ((rect->y + t) * tw + (rect->x + s)) + 1] = (unsigned char) (255.0f * light);
					}
				}

				index++;
			}
		}
		*/

/*

		Image image;
		image.loadFromMemory(lMap, tw, th, FORMAT_I8, true, false);
//		image.loadFromMemory(lMap, tw, th, FORMAT_IA8, true, false);
		char str[64];
		sprintf(str, "LightMap%d.dds", lm);
		image.saveImage(str);
	}
*/
	return true;
}

bool MainApp::exit(){
	delete model;
	return true;
}

bool MainApp::load(){
	if (!GL_ARB_shader_objects_supported || !GL_ARB_vertex_shader_supported || !GL_ARB_fragment_shader_supported || !GL_ARB_shading_language_100_supported){
		addToLog("No GLSL support (GL_ARB_shader_objects, GL_ARB_vertex_shader, GL_ARB_fragment_shader, GL_ARB_shading_language_100)\n");
		return false;
	}

	if ((lighting = renderer->addShader("lighting.shd")) == SHADER_NONE) return false;

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	if ((light = renderer->addTexture("../Textures/Particle.png", TEX_CLAMP)) == TEXTURE_NONE) return false;

//	if ((base[0] = renderer->addTexture("testIA8.dds")) == TEXTURE_NONE) return false;
	if ((base[0] = renderer->addTexture("../Textures/woodfloor1.dds")) == TEXTURE_NONE) return false;
	if ((bump[0] = renderer->addTexture("../Textures/woodfloor1Bump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;

	if ((base[1] = renderer->addTexture("../Textures/stone34.dds")) == TEXTURE_NONE) return false;
	if ((bump[1] = renderer->addTexture("../Textures/stone34Bump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;
	
	if ((base[2] = renderer->addTexture("../Textures/Tx_imp_wall_01_small.dds")) == TEXTURE_NONE) return false;
	if ((bump[2] = renderer->addTexture("../Textures/Tx_imp_wall_01Bump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;
	
	if ((base[3] = renderer->addTexture("../Textures/StoneWall_1-2.dds")) == TEXTURE_NONE) return false;
	if ((bump[3] = renderer->addTexture("../Textures/StoneWall_1-2Bump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;

	
	for (int i = 0; i < LIGHTMAP_COUNT; i++){
		char str[64];
		sprintf(str, "LightMap%d.dds", i);
		if ((lightMaps[i] = renderer->addTexture(str, TEX_NOMIPMAPPING)) == TEXTURE_NONE) return false;
	}

	model->uploadToVertexBuffer();

	return true;
}

bool MainApp::unload(){
	model->freeVertexBuffer();
	return true;
}

void MainApp::drawLight(){
	renderer->setMask(COLOR);
	renderer->apply();

	glColor3f(0, 0, 0);

	vec3 rPos = vec3(256, 416, 0);
	vec3 dir0 = lightPos - rPos;
	vec3 dir1 = normalize(dir0);
	dir1 = vec3(0, 8 * dir1.z, -8 * dir1.y);
	vec3 dir2(8, 0, 0);

	glBegin(GL_TRIANGLE_STRIP);
		glVertex3fv(rPos - dir1 - dir2);
		glVertex3fv(rPos + dir0 - dir1 - dir2);
		glVertex3fv(rPos + dir1 - dir2);
		glVertex3fv(rPos + dir0 + dir1 - dir2);
		glVertex3fv(rPos + dir1 + dir2);
		glVertex3fv(rPos + dir0 + dir1 + dir2);
		glVertex3fv(rPos - dir1 + dir2);
		glVertex3fv(rPos + dir0 - dir1 + dir2);
		glVertex3fv(rPos - dir1 - dir2);
		glVertex3fv(rPos + dir0 - dir1 - dir2);
	glEnd();

	glColor3f(1, 1, 1);

	renderer->setBlending(GL_ONE, GL_ONE);
	renderer->setTextures(light);
	renderer->setMask(COLOR);
	renderer->apply();

	vec3 dx(modelView.elem[0][0], modelView.elem[0][1], modelView.elem[0][2]);
	vec3 dy(modelView.elem[1][0], modelView.elem[1][1], modelView.elem[1][2]);
	glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3fv(lightPos - lightSize * dx + lightSize * dy);

		glTexCoord2f(1,0);
		glVertex3fv(lightPos + lightSize * dx + lightSize * dy);

		glTexCoord2f(1,1);
		glVertex3fv(lightPos + lightSize * dx - lightSize * dy);

		glTexCoord2f(0,1);
		glVertex3fv(lightPos - lightSize * dx - lightSize * dy);
	glEnd();
}

bool MainApp::drawFrame(){
	renderer->changeMask(ALL);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	projection = projectionMatrixX(1.5f, float(height) / float(width), 0.1f, 6000);
	glLoadMatrixf(transpose(projection));

	modelView = rotateZXY(-wx, -wy, -wz);
	modelView.translate(-position);

    glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(transpose(modelView));


	float zp = sinf(1.5f * time);

	lightPos = getLightPos(zp);

	zp = (LIGHTMAP_COUNT - 1) * 0.5f * (zp + 1);
	int zi = int(zp);
	zp = fmodf(zp, 1.0f);
	

	glEnable(GL_CULL_FACE);

	for (unsigned int i = 0; i < model->getBatchCount(); i++){
		renderer->setShader(lighting);
		renderer->setTexture("Base", base[i]);
		renderer->setTexture("Bump", bump[i]);
		renderer->setTexture("LightMap0", lightMaps[min(zi, LIGHTMAP_COUNT - 1)]);
		renderer->setTexture("LightMap1", lightMaps[min(zi + 1, LIGHTMAP_COUNT - 1)]);

		renderer->apply();

		renderer->changeShaderConstant1f("weight", zp);
		renderer->changeShaderConstant3f("lightPos", lightPos);
		renderer->changeShaderConstant3f("camPos", position);

		((OpenGLBatch *) model->getBatch(i))->draw();
	}

	drawLight();

	return true;
}

App *app = new MainApp();
